from .big_bench_hard.big_bench_hard import BigBenchHard
from .mmlu.mmlu import MMLU
from .hellaswag.hellaswag import HellaSwag
from .drop.drop import DROP
from .truthful_qa.truthful_qa import TruthfulQA
from .gsm8k.gsm8k import GSM8K
from .human_eval.human_eval import HumanEval
